#include "libcell.h"

void write_object_value (_comm_header *comm, _path *path, _tag_detail *tag, int debug)
{
unsigned long a,b;
int x,size,offset,totalsize;
_tag_detail *basetag;
_data_buffer *buff;
_encaps_header *head;
_data_buffer *cpf_buffer;
_data_buffer *sendbuff;
_data_buffer *msgbuff;
_ioi_data *ioi1;
_ioi_data *ioi2;

dprint (DEBUG_TRACE,"write_object_value.c entered.\n");
ioi2 = NULL;

if (path == NULL)
	{
	printf ("path structure not allocated - write_object_value.c\n");
	return;
	}

if (tag == NULL)
	{
	printf ("tag structure not allocated - write_object_value.c\n");
	return;
	}

buff = malloc(sizeof(_data_buffer));
if (buff == NULL)
	{
	printf ("Could not allocate memory for data buffer - write_object_value.c\n");
	return;
	}
msgbuff = malloc(sizeof(_data_buffer));
if (msgbuff == NULL)
	{
	printf ("Could not allocate memory for msgdata buffer - write_object_value.c\n");
	free (buff);
	return;
	}
head = malloc(sizeof(_encaps_header));
if (head == NULL)
	{
	printf ("Could not allocate memory for Encaps Header - write_object_value.c\n");
	free (buff);
	free (msgbuff);
	return;
	}
cpf_buffer = malloc(sizeof(_data_buffer));
if (cpf_buffer == NULL)
	{
	printf ("Could not allocate memory for cpf_buffer - write_object_value.c\n");
	free(buff);
	free(head);
	free(msgbuff);
	return;
	}
sendbuff = malloc(DATA_Buffer_Length);
if (sendbuff == NULL)
	{
	printf ("Could not allocate memory for send buffer - write_object_value.c\n");
	free(buff);
	free(head);
	free(msgbuff);
	free(cpf_buffer);
	return;
	}
ioi1 = malloc(sizeof(_ioi_data));
if (ioi1 == NULL)
	{
	printf ("Could not allocate memory for ioi1 buffer - write_object_value.c\n");
	free (buff);
	free (head);
	free (msgbuff);
	free (cpf_buffer);
	free (sendbuff);
	return;
	}
basetag = NULL;

if (tag->alias_size != 0)
	{
	/* This is where we attempt to handle writing aliases.  It is not as clean as I would have liked it
	to be, and it probably won't work in all circumstances as it is written, but it should cover the 
	important aspects.  This is particularily designed to handle aliases pointing into an array of
	objects (i.e. an IO container).  Other cases may or may not work. 
	
	Please note the importance of how this is executed.  We must read the entire IO container first, 
	then we modify the relevent data within the container, then we write the data back out.  This must
	take place as quickly as possible to minimize the effect of the write routine stepping on values that
	may have been updated by the PLC in the background. */
	
	
	basetag = malloc (sizeof (_tag_detail));
	if (basetag == NULL)
		{
		printf ("Could not allocate memory for temporary tag - write_object_value.c\n");
		return;
		}
	memset (basetag, 0 , sizeof(_tag_detail));
	basetag->topbase = tag->alias_topbase;
	basetag->base = tag->alias_base;
	basetag->id = tag->alias_id;
	basetag->linkid = tag->alias_linkid;
	basetag->type = tag->alias_type;
	basetag->size = tag->alias_size;
	basetag->data = malloc (getsize(basetag));
	read_object_value (comm, path, basetag, debug);
	if (debug != DEBUG_NIL)
		{
		dprint (DEBUG_VALUES, "Initial data for alias write: ");
		for (x=0; x<basetag->datalen; x++)
			dprint (DEBUG_VALUES, "%02X ",basetag->data[x]);
		dprint (DEBUG_VALUES, "\n");
		}
	memcpy (basetag->data + (tag->linkid - tag->alias_linkid), tag->data, tag->size);
	if (debug != DEBUG_NIL)
		{
		dprint (DEBUG_VALUES, "Data to be written: ");
		for (x=0; x<basetag->datalen; x++)
			dprint (DEBUG_VALUES, "%02X ",basetag->data[x]);
		dprint (DEBUG_VALUES, "\n");
		}
	write_object_value (comm, path, basetag, debug);
	free (basetag);
	free (ioi1);
	free (buff);
	free (head);
	free (msgbuff);
	free (cpf_buffer);
	free (sendbuff);
	return;
	}
	

totalsize=getsize(tag);
size=0;
offset=0;

while (1)
{	
size = totalsize-offset;
if (size > 255)
	size=255;
dprint (DEBUG_TRACE,"Write loop entered - totalsize = %d, size = %d, offset = %d\n",totalsize,size,offset);
memset(buff,0,sizeof(_data_buffer));
memset(msgbuff,0,sizeof(_data_buffer));
memset(head,0,sizeof(_encaps_header));
memset(cpf_buffer,0,sizeof(_data_buffer));
memset(sendbuff,0,sizeof(_data_buffer));
memset(ioi1,0,sizeof(_ioi_data));


buff->len = 0;

msgbuff->data[msgbuff->len++] = PUT_SINGLE_PROP;
ioi1->class = OBJECT_CONFIG;
ioi1->instance = tag->topbase;
if (tag->alias_topbase != 0)
	ioi1->instance = tag->alias_topbase;

ioi1->member = -1;
ioi1->point = -1;
ioi1->attribute = -1;
ioi1->tagname = NULL;
a = tag->id;
b = tag->topbase;
if (tag->alias_id != 0)
	a = tag->alias_id;
if (tag->alias_topbase != 0)
	b = tag->alias_topbase;	
if ((a != 0) && (a == b))
	{
	ioi1->class = OBJECT_SUB_OBJECT;
	ioi1->instance = tag->base;
	if (tag->alias_base != 0)
		ioi1->instance = tag->alias_base;
	ioi2 = malloc(sizeof(_ioi_data));
	if (ioi2 == NULL)
		{
		printf ("Could not allocate memory for ioi2 buffer - write_object_data.c\n");
		free (buff);
		free (head);
		free (msgbuff);
		free (cpf_buffer);
		free (sendbuff);
		free (ioi1);
		}
	memset(ioi2,0,sizeof(_ioi_data));
	ioi2->class = OBJECT_CONFIG;
	ioi2->instance = tag->id;
	if (tag->alias_id != 0)
		ioi2->instance = tag->alias_id;
	ioi2->member = -1;
	ioi2->point = -1;
	ioi2->attribute = -1;
	ioi2->tagname = NULL;
	}
ioi(msgbuff, ioi1, ioi2 ,debug);
msgbuff->data[msgbuff->len++] = (offset & 255);
msgbuff->data[msgbuff->len++] = (offset / 0x100);
msgbuff->data[msgbuff->len++] = 0;
msgbuff->data[msgbuff->len++] = 0;
msgbuff->data[msgbuff->len++] = 0;
msgbuff->data[msgbuff->len++] = size;
msgbuff->data[msgbuff->len++] = 0;
memcpy (msgbuff->data + msgbuff->len, tag->data + offset, size);
msgbuff->len += size;
if ((size & 1) == 0)
	msgbuff->data[msgbuff->len++] = 0;
//msgbuff->data[msgbuff->len++] = 0;

if (debug != DEBUG_NIL)
	{
	dprint (DEBUG_BUILD,"Msgbuff = ");
	for (x=0;x<msgbuff->len;x++)
		dprint (DEBUG_BUILD,"%02X ",msgbuff->data[x]);
	dprint (DEBUG_BUILD,"\n");
	}

cpf_buffer->data[cpf_buffer->len++] = PDU_Unconnected_Send;
ioi1->class = CONNECTION_MANAGER;
ioi1->instance = FIRST_INSTANCE;
ioi1->member = -1;
ioi1->point = -1;
ioi1->attribute = -1;
ioi1->tagname = NULL;

ioi (cpf_buffer, ioi1, NULL, debug);
settimeout (6,0x9a,cpf_buffer, debug);
cpf_buffer->data[cpf_buffer->len++] = msgbuff->len & 255;
cpf_buffer->data[cpf_buffer->len++] = msgbuff->len / 0x100;

makepath(path, msgbuff, debug);
memcpy (cpf_buffer->data + cpf_buffer->len, msgbuff->data, msgbuff->len);
cpf_buffer->len += msgbuff->len;

sendRRdata (10, comm, head, buff, debug);
cpf (CPH_Null, NULL, CPH_Unconnected_message, cpf_buffer, buff, debug);

head->len = buff->len;
memcpy (sendbuff->data, head, ENCAPS_Header_Length);
sendbuff->overall_len = ENCAPS_Header_Length;
if (debug!= DEBUG_NIL)
	{
	dprint (DEBUG_BUILD,"Loading sendbuffer with command data.\n");
	for (x=0;x<buff->len;x++)
		dprint (DEBUG_BUILD,"%02X ",buff->data[x]);
	dprint (DEBUG_BUILD,"\n");
	}
memcpy (sendbuff->data + sendbuff->overall_len, buff->data, buff->len);
sendbuff->overall_len += buff->len;
senddata (sendbuff, comm, debug);
readdata (buff, comm, debug);
memset (head,0, ENCAPS_Header_Length);
memcpy (head,buff->data,ENCAPS_Header_Length);
if (head->status != 0)
	{
	dprint (DEBUG_VALUES,"Write Object Value command returned an error.\n");
	if (ioi2 != NULL)
		free (ioi2);
	free(ioi1);
	free(buff);
	free(head);
	free(cpf_buffer);
	free(sendbuff);
	free(msgbuff);
	return;
	}
//buff = buff + ENCAPS_Header_Length;
if (debug != DEBUG_NIL)
	{
	dprint (DEBUG_VALUES,"Got good reply to Write Object Value Command - %d\n",buff->overall_len);	
	for (x=44;x < buff->overall_len;x++)
		dprint (DEBUG_VALUES,"%02X ",buff->data[x]);
	dprint (DEBUG_VALUES,"\n");
	}
if (size+offset >= totalsize)
	break;
offset += size;
}	

if (ioi2 != NULL)
	free (ioi2);
free(ioi1);
free(sendbuff);
free(cpf_buffer);
free(head);
free(msgbuff);
free(buff);
dprint (DEBUG_TRACE,"write_object_value.c exited.\n");
return;
}
